//+------------------------------------------------------------------+
//|                                                           MemMap |
//|                                 Copyright 2006-2012, FXmaster.de |
//|       programming & support - Alex Sergeev (profy.mql@gmail.com) |
//+------------------------------------------------------------------+
#property copyright "Copyright 2006-2012, Alex Sergeev"
#property link      "www.FXmaster.de"

//     WinApi

#define BYTE		int
#define DWORD		int
#define BOOL		int
#define LPTSTR		string
#define PBYTE		int
#define LPCTSTR	const string
#define LPVOID		int
#define LPCVOID 	const int
#define SIZE_T		int
#define HANDLE		int
#define WINAPI		
#define FALSE		0

#define INVALID_HANDLE_VALUE ((HANDLE)(int)-1)
#define PAGE_READWRITE						0x04     
#define FILE_MAP_ALL_ACCESS 				SECTION_ALL_ACCESS
#define STANDARD_RIGHTS_REQUIRED			(0x000F0000)
#define SECTION_QUERY               	0x0001
#define SECTION_MAP_WRITE           	0x0002
#define SECTION_MAP_READ            	0x0004
#define SECTION_MAP_EXECUTE         	0x0008
#define SECTION_EXTEND_SIZE         	0x0010
#define SECTION_MAP_EXECUTE_EXPLICIT	0x0020 // not included in SECTION_ALL_ACCESS
#define SECTION_ALL_ACCESS					(STANDARD_RIGHTS_REQUIRED|SECTION_QUERY|SECTION_MAP_WRITE|SECTION_MAP_READ|SECTION_MAP_EXECUTE|SECTION_EXTEND_SIZE)
//+------------------------------------------------------------------+
//| SECURITY_ATTRIBUTES                                              |
//+------------------------------------------------------------------+
struct SECURITY_ATTRIBUTES
  {
   DWORD             nLength;
   LPVOID            lpSecurityDescriptor;
   BOOL              bInheritHandle;
  };

#define LPSECURITY_ATTRIBUTES			int

#import "kernel32.dll"
HANDLE WINAPI OpenFileMappingW(DWORD dwDesiredAccess,BOOL bInheritHandle,LPCTSTR lpName);
HANDLE WINAPI CreateFileMappingW(HANDLE hFile,LPSECURITY_ATTRIBUTES lpAttributes,DWORD flProtect,DWORD dwMaximumSizeHigh,DWORD dwMaximumSizeLow,LPCTSTR lpName);
LPVOID WINAPI MapViewOfFile(HANDLE hFileMappingObject,DWORD dwDesiredAccess,DWORD dwFileOffsetHigh,DWORD dwFileOffsetLow,SIZE_T dwNumberOfBytesToMap);
BOOL WINAPI UnmapViewOfFile(LPCVOID lpBaseAddress);
BOOL CloseHandle(HANDLE hObject);
#import "kernel32.dll"
int GetLastError();
#import "msvcrt.dll"
int memset(int &Destination[],int    c,       int Length);
int memset(int  Destination,  int    c,       int Length);
int memcpy(int &Destination[],int   &Source[],int Length);
int memcpy(int &Destination[],int    Source,  int Length);
int memcpy(int  Destination,  int   &Source[],int Length);
#import

#define OpenFileMapping			OpenFileMappingW
#define CreateFileMapping		CreateFileMappingW


#define HEAD_MEM		4 //   ,    
enum OpenFlags { modeOpen=(int) 0x00000,modeCreate=(int)0x00001 }; // Flag values
//---      Mapping
//+------------------------------------------------------------------+
//| class CMemMapAPI                                                 |
//+------------------------------------------------------------------+
class CMemMapApi
  {
public:
                     CMemMapApi() { };
                    ~CMemMapApi() { };
protected:
   virtual HANDLE    Open(LPTSTR path,DWORD size,int mode,DWORD &err);                  // 
   virtual void      Close(HANDLE hmem);                                                // 
   virtual int       Fill(HANDLE h,BYTE b,DWORD &err);                                  //    
   virtual HANDLE    Grows(HANDLE hmem,LPTSTR path,DWORD size,DWORD &err);              //  
   virtual PBYTE     ViewFile(HANDLE hmem,DWORD &err);                                  //  
   virtual void      UnViewFile(PBYTE buf);                                             //  
   virtual DWORD     GetSize(HANDLE hmem,DWORD &err);                                   //  
   virtual int       SetSize(HANDLE hmem,DWORD size,DWORD &err);                        //  
   virtual int       Write(HANDLE hmem,const int &buf[],DWORD pos,int sz,DWORD &err); //           
   virtual int       Read(HANDLE hmem,int &buf[],DWORD pos,int sz,DWORD &err);        //           
  };
//+------------------------------------------------------------------+
//| Open                                                             |
//+------------------------------------------------------------------+
HANDLE CMemMapApi::Open(LPTSTR path,DWORD size,int mode,DWORD &err)
  {
   err=0;
   if(path=="") return(NULL);
   HANDLE hmem=NULL;
   if(mode==modeCreate) hmem=CreateFileMapping(INVALID_HANDLE_VALUE,NULL,PAGE_READWRITE,0,size+HEAD_MEM,path); //   
   if(mode==modeOpen) hmem=OpenFileMapping(FILE_MAP_ALL_ACCESS,FALSE,path); //   
   if(hmem==NULL) { err=kernel32::GetLastError(); return(NULL); }//   
   if(mode==modeCreate) //   ,   
     {
      DWORD r=SetSize(hmem,size,err); if(r!=0 || err!=0) { Close(hmem); return(NULL); }
     }
   return(hmem);
  }
//+------------------------------------------------------------------+
//| Close                                                            |
//+------------------------------------------------------------------+
void CMemMapApi::Close(HANDLE hmem)
  {
   if(hmem!=NULL) CloseHandle(hmem); hmem=NULL; //  
  }
//+------------------------------------------------------------------+
//| Fill                                                             |
//+------------------------------------------------------------------+
int CMemMapApi::Fill(HANDLE hmem,BYTE b,DWORD &err) //    
  {
   if(hmem==NULL) return(0);
   PBYTE view=ViewFile(hmem,err); if(view==0 || err!=0) return(-1); //   
   DWORD size=GetSize(hmem, err); if(size<=0 || err!=0) return(-2); //  
   memset(view,b,size);
   return(size);
  }
//+------------------------------------------------------------------+
//| Grows                                                            |
//+------------------------------------------------------------------+
HANDLE CMemMapApi::Grows(HANDLE hmem,LPTSTR path,DWORD newsize,DWORD &err)
  {
   if(hmem==NULL) { err=-1; return(0); } //   
   DWORD size=GetSize(hmem,err); if(newsize<=size || err!=0) return(hmem); //  
   HANDLE hnew=Open(path,newsize,modeCreate,err); if(hnew==NULL || err!=0) { CloseHandle(hnew); return(0); } //   
   CloseHandle(hmem); //  
   return(hnew); //  
  }
struct _byte4 { int b[4]; };
struct _int { int b; };
//+------------------------------------------------------------------+
//| GetSize                                                          |
//+------------------------------------------------------------------+
DWORD CMemMapApi::GetSize(HANDLE hmem,DWORD &err)
  {
   PBYTE view=ViewFile(hmem,err); if(view==0 || err!=0) return(-1); //  
   int sz=sizeof(DWORD);
   _byte4 dest; //    
   memcpy(dest.b,view,sz);
   _int _size=dest;
   UnViewFile(view); //  
   return(_size.b); //  
  }
//+------------------------------------------------------------------+
//| SetSize                                                          |
//+------------------------------------------------------------------+
int CMemMapApi::SetSize(HANDLE hmem,DWORD size,DWORD &err)
  {
   PBYTE view=ViewFile(hmem,err); if(view==0 || err!=0) return(-1); //  
   int sz=sizeof(DWORD);
   _int _size; _size.b=size;
   _byte4 src=_size; //    
   memcpy(view,src.b,sz);
   UnViewFile(view); //  
   return(0); //  
  }
//+------------------------------------------------------------------+
//| ViewFile                                                         |
//+------------------------------------------------------------------+
PBYTE   CMemMapApi::ViewFile(HANDLE hmem,DWORD &err) //  
  {
   err=0;
   if(hmem==NULL) { err=-1; return(NULL); }//   
   PBYTE view=(PBYTE)MapViewOfFile(hmem,FILE_MAP_ALL_ACCESS,0,0,0); //   
   if(view==NULL) { err=kernel32::GetLastError(); return(NULL); } //   
   return(view); //     
  }
//+------------------------------------------------------------------+
//| UnViewFile                                                       |
//+------------------------------------------------------------------+
void CMemMapApi::UnViewFile(PBYTE view) //  
  {
   if(view!=NULL) UnmapViewOfFile(view); view=NULL; //  
  }
//+------------------------------------------------------------------+
//| Write                                                            |
//+------------------------------------------------------------------+
int CMemMapApi::Write(HANDLE hmem,const int &buf[],DWORD pos,int sz,DWORD &err) //      
  {
   
   if(hmem==NULL) return(-1);
   PBYTE view=ViewFile(hmem,err); if(view==0 || err!=0) return(-1); //   
   DWORD size=GetSize(hmem,err); if(pos+sz>size) { UnViewFile(view); return(-2); }; //   ,  
   int src[]; ArrayResize(src,size+HEAD_MEM); memcpy(src,view,size); //  
   
   for(int i=0; i<sz/4; i++) src[pos+i+HEAD_MEM]=buf[i]; //   
   
   memcpy(view,src,size); //  
   UnViewFile(view); //  
   return(0); //  
  }
//+------------------------------------------------------------------+
//| Read                                                             |
//+------------------------------------------------------------------+
int CMemMapApi::Read(HANDLE hmem,int &buf[],DWORD pos,int sz,DWORD &err) //      
  {
   if(hmem==NULL) return(-1);
   PBYTE view=ViewFile(hmem,err); if(view==0 || err!=0) return(-1); //   
   DWORD size=GetSize(hmem,err); //  
   int src[]; ArrayResize(src,size); memcpy(src,view,size); //  
   ArrayResize(buf,sz);
   int i=0; for(i=0; i<sz/4 && pos+i<size/4; i++) buf[i]=src[pos+i+HEAD_MEM]; //  
   UnViewFile(view); //   
   return(i); //   
  }
//+------------------------------------------------------------------+
//| class CMemMapApi                                                 |
//+------------------------------------------------------------------+
class CMemMapFile: public CMemMapApi
  {
public:
   HANDLE            m_hmem;  // 
   LPTSTR            m_path;  //   
   DWORD             m_size;  //  
   DWORD             m_pos;   //   
   int               m_mode;  //  
                              // PBYTE m_buf; //    
   DWORD             err;
public:
                     CMemMapFile();
                    ~CMemMapFile();
public:
   virtual HANDLE    Open(LPTSTR path,DWORD size,int mode);         // 
   virtual void      Close();                                       //    
   virtual int       Fill(BYTE b);                                  //    
   virtual int       Seek(DWORD pos,int orig);                      //     
   virtual int       Grows(DWORD size);                             //  
   virtual int IsEOF() { if(m_pos>=m_size) return(1); return(0); }; // 
   virtual DWORD Tell() { return(m_pos); };                         // 
   virtual int       Write(const int &buf[],int sz);              //      
   virtual int       Read(int &buf[],int sz);                     //      
  };
//+------------------------------------------------------------------+
//| CMemMapFile class constructor                                    |
//+------------------------------------------------------------------+
CMemMapFile::CMemMapFile()
  {
   m_path=""; m_hmem=NULL; m_size=0; m_pos=0; m_mode=-1;
  }
//+------------------------------------------------------------------+
//| CMemMapFile class desstructor                                    |
//+------------------------------------------------------------------+
CMemMapFile::~CMemMapFile()
  {
   Close();
  }
//+------------------------------------------------------------------+
//| Open                                                             |
//+------------------------------------------------------------------+
int CMemMapFile::Open(LPTSTR path,DWORD size,int mode=modeOpen)
  {
   m_size=size; m_path=path; m_mode=mode; m_pos=0; //  
   if(m_path=="") return(-1);
   m_hmem=Open(m_path,size,mode,err);
   if(m_hmem==NULL) return(err); //   
   return(0);
  }
//+------------------------------------------------------------------+
//| Close                                                            |
//+------------------------------------------------------------------+
void CMemMapFile::Close()
  {
   if(m_hmem!=NULL) CloseHandle(m_hmem); m_path=""; m_hmem=NULL; m_size=0; m_pos=0; m_mode=-1; //  
  }
//+------------------------------------------------------------------+
//| Fill                                                            |
//+------------------------------------------------------------------+
int CMemMapFile::Fill(BYTE b) //    
  {
   if(m_hmem==NULL) return(-1); //   
   return(Fill(m_hmem,b,err));
  }
//+------------------------------------------------------------------+
//| Grows                                                            |
//+------------------------------------------------------------------+
int CMemMapFile::Grows(DWORD size)
  {
   if(m_hmem==NULL || m_path=="" || size<=0) return(-1);
   if(size<=m_size) return(0);
   HANDLE hnew=Grows(m_hmem,m_path,size,err);
   if(hnew==NULL) return(err); //   
   m_hmem=hnew; m_size=size;
   return(0);
  }
//+------------------------------------------------------------------+
//| Seek                                                             |
//+------------------------------------------------------------------+
int CMemMapFile::Seek(DWORD pos,int seek=SEEK_SET) //     
  {
   if(seek==SEEK_SET) m_pos=pos;
   if(seek==SEEK_CUR) m_pos+=pos;
   if(seek==SEEK_END) m_pos=m_size-pos;
// 
   m_pos=(m_pos<0)?0:m_pos;
   m_pos=(m_pos>m_size)?m_size:m_pos;
   return(0);
  }
//+------------------------------------------------------------------+
//| Write                                                            |
//+------------------------------------------------------------------+
int CMemMapFile::Write(const int &buf[],int sz) //      
  {
   
   if(m_hmem==NULL) return(-1); //   
   
   if(m_pos+sz>m_size) if(Grows(m_pos+sz)!=0) return(-2);
   
   Write(m_hmem,buf,m_pos,sz,err);
   
   m_pos+=sz; //  
   return(0);
  }
//+------------------------------------------------------------------+
//| Read                                                             |
//+------------------------------------------------------------------+
int CMemMapFile::Read(int &buf[],int sz) //      
  {
   if(m_hmem==NULL) return(-1); //   
   int i=Read(m_hmem,buf,m_pos,sz,err);
   m_pos+=i;  //  
   return(i); //     
  }
//+------------------------------------------------------------------+
